#include "PointStruct.h"

#include <stdio.h>
#include <stdlib.h>
#include <algorithm>

/*
 * Compares the u coordinate of the Point
 * @return 1 (true) if left.u_coord is smaller than right.u_coord
 *         0 (false) otherwise
 */
bool Point_sorter_u(Point const& left, Point const& right){
    return left.u_coord < right.u_coord;
}

/*
 * Compares the v coordinate of the Point
 * @return 1 (true) if left.v_coord is smaller than right.v_coord
 *         0 (false) otherwise
 */
bool Point_sorter_v(Point const& left, Point const& right){
    return left.v_coord < right.v_coord;
}



/*
 * Prints to screen the contents of a list
 * @param list is pointer to the start of list
 * @param listSize is how many values to printed out
 * @complexity O(listSize)
 */
void displayList(Point* list, int listSize){
    for (int i = 0; i < listSize; i++){
        printf("%f %f\n", list[i].u_coord,list[i].v_coord);
    }
}


/*
 * Duplicates an Point array. Leaves the original array unedited
 * @param list
 * @param listSize
 * @return pointer to the duplicate. It will have size listSize;
 * @complexity O(listSize)
 */
Point* duplicateList(Point*list, int listSize){
    Point *newArray = (Point*)malloc(listSize*sizeof(Point));
    for (int i = 0; i < listSize; i++){
        newArray[i].u_coord = list[i].u_coord;
        newArray[i].v_coord = list[i].v_coord;
    }
    return newArray;

}


/*
* Sorts the given list from smallest to largest
* uses the standard sort algorithm
* @param list
* @param listSize
* @param coord 0 for u_coord, 1 for v_coord
* @return pointer to the sorted list (different list to the original). 
*         It will have size listSize.
* @complexity O(listSize) + O(sort)
*/
Point* sortList(Point*list, int listSize, int coord){
    Point *duplicate = duplicateList(list, listSize);
    if (coord == 0){
        std::sort(duplicate, duplicate + listSize, &Point_sorter_u);
    }
    else if (coord == 1){
        std::sort(duplicate, duplicate + listSize, &Point_sorter_v);
    }
    else{
        printf("Error in Point* sortList.\n");
        exit(1);
    }
    
    return duplicate;
}

/*
 * Returns the median value of a sorted list
 *
 * @param list the list must be sorted
 * @param listSize
 * @return the median
 *         (for even sized lists, it returns the higher of the two center values)
 * @complexity: O(1)
 */
Point getMedian(Point *list, int listSize){
    int midpoint = listSize / 2;
    return *(list + midpoint);
}

/*
* 95% interval
*
* @param list the list must be sorted
* @param listSize
* @return a list of size 2 with the values at positions 2.5% and 97.5% (central 95%)
* @complexity: O(1)
*/
Point* get95Int(Point *list, int listSize){
    Point *Interval = (Point*)malloc(2 * sizeof(Point));
    int lower = (int)(listSize*0.025);
    int upper = (int)(listSize*0.975);
    *(Interval) = *(list + lower);
    *(Interval + 1) = *(list + upper);
    return Interval;
}


/*
 *
 * @param list. Expected that list is not sorted
 * @param listSize
 * @param whichcoord 0 for u_coord, 1 for v_coord
 * @return maximum value of list
 *
 * @post the list is unchanged.
 */
Point getMaxValue(Point *list, int listSize, int whichcoord){
    Point MaxValue = { list[0].u_coord, list[0].v_coord };
    if (whichcoord == 0){
        for (int i = 1; i < listSize; i++){
            if (list[i].u_coord>MaxValue.u_coord){
                MaxValue.u_coord = list[i].u_coord;
                MaxValue.v_coord = list[i].v_coord;
            }
        }
    }
    else if (whichcoord == 1){
        for (int i = 1; i < listSize; i++){
            if (list[i].v_coord>MaxValue.v_coord){
                MaxValue.u_coord = list[i].u_coord;
                MaxValue.v_coord = list[i].v_coord;
            }
        }
    }
    else{
        printf("Error in Point getMaxValue,\n");
        exit(1);
    }

    return MaxValue;
}

/*
 *
 * @param list. Expected that list is not sorted
 * @param listSize
 * @param whichcoord 0 for u_coord, 1 for v_coord
 * @return minimum value of list
 *
 * @post the list is unchanged.
 */
Point getMinValue(Point *list, int listSize, int whichcoord){
    Point MinValue = { list[0].u_coord, list[0].v_coord };
    if (whichcoord == 0){
        for (int i = 1; i < listSize; i++){
            if (list[i].u_coord<MinValue.u_coord){
                MinValue.u_coord = list[i].u_coord;
                MinValue.v_coord = list[i].v_coord;
            }
        }
    }
    else if (whichcoord == 1){
        for (int i = 1; i < listSize; i++){
            if (list[i].v_coord<MinValue.v_coord){
                MinValue.u_coord = list[i].u_coord;
                MinValue.v_coord = list[i].v_coord;
            }
        }
    }
    else{
        printf("Error in Point getMinValue,\n");
        exit(1);
    }
    
    return MinValue;
}




