#include <stdio.h>
#include <stdlib.h>
#include "ReadData.h"

/*
 * displays Location list
 */
void displayListOfLocations(Location *LocationList, int listSize){
    printf("x y timestart timeend\n");
    for (int i = 0; i < listSize; i++){
        printf("%d: %f %f %d %d\n",
            i, LocationList[i].x_coord, LocationList[i].y_coord,
            LocationList[i].time_start, LocationList[i].time_end);
    }
}


/*
 * Reads in the file containing details about the observed nests
 *
 * @complexity ??
 */
void readFixed(Location *LocationList){
    char fixedCoordsFileName[] = "fixed";

    FILE *fixedCoords = fopen(fixedCoordsFileName, "r");
    if (!fixedCoords){
        printf("Could not open file %s.\n", fixedCoordsFileName);
        exit(1);
    }
    else{
        printf("Opened file %s.\n", fixedCoordsFileName);
    }

    int dummyDeathType;
    int tempX;
    int tempY;
    int tempTimeDeath;

    for (int i = 0; i < NUMOBSERVEDNESTS; i++){
        if (fscanf(fixedCoords, "%d %d %d %d",
            &tempX, &tempY, &tempTimeDeath, &dummyDeathType) == 4){
            //successful
            LocationList[i].x_coord = tempX;
            LocationList[i].y_coord = tempY;
            LocationList[i].time_end = tempTimeDeath;
            //printf("%f %f %d\n", LocationList[i].x_coord, LocationList[i].y_coord, LocationList[i].time_end);
            //LocationList[i].time_start = LocationList[i].time_end;//temporary. Remove
        }
        else{
            printf("Failed readFixed nests.\n");
            exit(1);
        }

    }
    fclose(fixedCoords);

}


/*
 * Reads in the sample number and total number of nests
 *
 * @return maxNumNests the number of nests that are relevant when accessing
 *         this particular sample
 */
int readSampleHeader(FILE *nestCoords){
    int sampleNum;
    int maxNumNests;
    if (fscanf(nestCoords, "%d. %d\n", &sampleNum, &maxNumNests) == 2){
        printf("%d. %d\n", sampleNum, maxNumNests);
        //successful
    }
    else{
        printf("Reading in sample 'header' failed");
        exit(1);
    }

    return maxNumNests;
}


/*
 * Reads in one sample from the file
 *
 * @param in the file for nest coordinates
 * @param NestList the pointer containing the data for this given sample
 * @complexity ??
 */
int readSample(FILE *nestCoords, int maxNumNests, Location* LocationList){
    int actualNumNests;

    for (actualNumNests = 0; actualNumNests < NUMOBSERVEDNESTS; actualNumNests++){

        if (fscanf(nestCoords, "%d",
            &(LocationList[actualNumNests].time_start)) == 1){
            //successful
            //printf("%d: %f %f %d %d\n", actualNumNests,
            //    LocationList[actualNumNests].x_coord,
            //    LocationList[actualNumNests].y_coord,
            //    LocationList[actualNumNests].time_start,
            //    LocationList[actualNumNests].time_end);

        }
        else{
            printf("Reading in sample observed nests time_start failed");
            exit(1);
        }
    }

    //at this point, actualNumNests = NUMOBSERVEDNESTS
    double tempX=-1.0;
    double tempY=-1.0;
    int tempTimeStart=9999;
    char tempTimeEnd[3] = { '0', '0', '0'};

    for (int i = actualNumNests; i < maxNumNests; i++){
        if (fscanf(nestCoords, "%lf %lf %d %s",
                    &tempX, &tempY, &tempTimeStart, tempTimeEnd) == 4){

            //printf("%lf %lf %d %s\n", tempX, tempY, tempTimeStart, tempTimeEnd);

            if (tempTimeEnd[0] == '&'){
                //then non established nest
                //do not include;
            }
            else{
                if (tempTimeEnd[0] == '*'){
                    LocationList[actualNumNests].time_end = 9999;
                }
                else{
                    LocationList[actualNumNests].time_end = atoi(tempTimeEnd);
                }
                LocationList[actualNumNests].x_coord = tempX + MINX;//
                LocationList[actualNumNests].y_coord = tempY + MINY;//
                LocationList[actualNumNests].time_start = tempTimeStart;
                //printf("%d: %f %f %d %d\n",
                //    actualNumNests,
                //    LocationList[actualNumNests].x_coord,
                //    LocationList[actualNumNests].y_coord,
                //    LocationList[actualNumNests].time_start,
                //    LocationList[actualNumNests].time_end);

                actualNumNests++;
            }
        }
        else{
            printf("Reading in sample unobserved nests failed");
            exit(1);
        }

    }

    return actualNumNests;

}

/*
 *  Goes through and reads ONE sample data, and ignores it (i.e. does nothing)
 */
void ignoreSample(FILE *nestCoords){
    int maxNumNests = readSampleHeader(nestCoords);
    int actualNumNests;
    int time_start_dummy=0;

    for (actualNumNests = 0; actualNumNests < NUMOBSERVEDNESTS; actualNumNests++){

        if (fscanf(nestCoords, "%d", &time_start_dummy) == 1){
            //successful
        }
        else{
            printf("Reading in sample observed nests time_start failed");
            exit(1);
        }
    }

    //at this point, actualNumNests = NUMOBSERVEDNESTS
    double tempX = -1.0;
    double tempY = -1.0;
    int tempTimeStart = 9999;
    char tempTimeEnd[3] = { '0', '0', '0' };

    for (int i = actualNumNests; i < maxNumNests; i++){
        if (fscanf(nestCoords, "%lf %lf %d %s",
            &tempX, &tempY, &tempTimeStart, tempTimeEnd) == 4){
            //printf("%lf %lf %d %s\n", tempX, tempY, tempTimeStart, tempTimeEnd);

        }
        else{
            printf("Reading in sample unobserved nests failed");
            exit(1);
        }

    }

}


