/*******************************************************************************
*
* ants_map
* Code written by Jonathan Keith starting with code written by Thao Le
* July 2015
* As a method to display boundaries of the fire ants spread
* Method: convexpolygon05 file (Thao)
*
********************************************************************************
*/

#include <iostream>
#include <string>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
//#include "boost/multi_array.hpp"

#include "listFunctions.h" //functions for 1D double arrays
#include "PointStruct.h" //for Point arrays
//#include "Array3D.h" //for 3D arrays implemented as 1D
#include "ReadData.h"

//#define XORIGIN 500000 //average (center) value of MINX_ORIG and MAXX_ORIG
//#define YORIGIN 6945000

//#define SAMPLESIZE 3 //change as relevant
//#define IGNORESAMPLE 19997 //change as relevant
#define MINLEN 10000
int SAMPLESIZE; // after burn in
int IGNORESAMPLE; //burn in


//number of directions along which the boundaries will be found,
int numDirections;

//angle increment, computed using numDirections
double angleIncrement;

char nestCoordsFileName[100];
char outputFileName[100];

/*
 * Defines a wrapper class for a 3D array, implemented as a 1D array
 */
template <typename T>
class Array3D{
    // private data members
private:
    int rowNum;
    int columnNum;
    int depthNum;
    T* Matrix;

public:
    /*
     * Constructor
     */
    Array3D(int row, int column, int depth){
        if (row > 0 && column > 0 && depth > 0){
            rowNum = row;
            columnNum = column;
            depthNum = depth;
            Matrix = new T[rowNum *columnNum*depthNum + columnNum*depthNum];
        }

    }

    /*
     * Deconstructor
     */
    ~Array3D(){
        delete[] Matrix;
    }

    /*
     * Returns pointer to item at Matrx[x][y][0]
     * i.e, an "array" of size depthNum
     */
    T* operator () (int x, int y){
        return &Matrix[x*depthNum*columnNum + y*depthNum + 0];
    }

    /*
     * Returns item at Matrix[x][y][z]
     */
    const T& operator () (int x, int y, int z) const{
        return Matrix[x*depthNum*columnNum + y*depthNum + z];
    }
    T& operator () (int x, int y, int z){
        return Matrix[x*depthNum*columnNum + y*depthNum + z];
    }

    int getRowNum(){
        return rowNum;
    }

    int getColumnNum(){
        return columnNum;
    }

    int getDepthNum(){
        return depthNum;
    }
};


/*
* Defines a wrapper class for a 2D array, implemented as a 1D array
*/
template <typename T>
class Array2D{
    // private data members
private:
    int rowNum;
    int columnNum;
    T* Matrix;

public:
    /*
    * Constructor
    */
    Array2D(int row, int column){
        if (row > 0 && column > 0){
            rowNum = row;
            columnNum = column;
            Matrix = new T[rowNum *columnNum + columnNum];
        }

    }

    /*
    * Deconstructor
    */
    ~Array2D(){
        delete[] Matrix;
    }

    /*
    * Returns pointer to item at Matrx[x][y][0]
    * i.e, an "array" of size depthNum
    */
    T* operator () (int x){
        return &Matrix[x*columnNum + 0];
    }

    /*
    * Returns item at Matrix[x][y][z]
    */
    const T& operator () (int x, int y) const{
        return Matrix[x*columnNum + y];
    }
    T& operator () (int x, int y){
        return Matrix[x*columnNum + y];
    }

    int getRowNum(){
        return rowNum;
    }

    int getColumnNum(){
        return columnNum;
    }

};


struct edge{
  int p1; //End point 1
  int p2; //End point 2
  int t1; //occurence of edge in indexlist
};


struct poly{
  int numpoints;
  float *points;
  double area;
};


/*
* @returns double array with values in order a, b, c,
*          where ax+by=c
*/
double* findEquationOfLine(double x1, double y1, double x2, double y2){
    //returns double
    double *array = (double*)malloc(3 * sizeof(double));
    array[0] = y1 - y2;
    array[1] = x2 - x1;
    array[2] = x2*y1 - x1*y2;
    return array;


}

double* findEquationOfLine(Point first, Point second){
    //returns double
    double *array = (double*)malloc(3 * sizeof(double));
    array[0] = first.v_coord - second.v_coord;
    array[1] = second.u_coord - first.u_coord;
    array[2] = second.u_coord*first.v_coord - first.u_coord*second.v_coord;
    return array;

}

/*
* employing Cramers' Rule
* @param line1 = {a,b,c} where ax+by=c
*/
Point findIntersectionOfLines(double *line1, double*line2){
    //a1*b2 - a2*b1
    double det = line1[0] * line2[1] - line2[0] * line1[1];
    double xnumerator = line2[1] * line1[2] - line1[1] * line2[2];
    double ynumerator = line1[0] * line2[2] - line2[0] * line1[2];
    Point returnpt = { xnumerator / det, ynumerator / det };
    return returnpt;

}

/*
 * application on cramer's rule and rotation matrix
 */
double findIntersectionParametric(double m1, double theta1, double m2, double theta2){
    return (m1*cos(theta1 - theta2) - m2) / sin(theta1 - theta2);
}


Point pointOnLine(double m, double theta, double t){
    Point coord = { m*cos(theta) - t*sin(theta), m*sin(theta) + t*cos(theta) };
    return coord;
}

double paramGivenPoint(double m, double theta, Point coord){
    double param;
    if (fabs(theta) <= 0.0001 || fabs(theta - PI) <= 0.0001){
        //can't use sintheta because sintheta=0 on denominator is bad
        param = (coord.v_coord - m*sin(theta)) / cos(theta);

    }
    else{
        param = (m*cos(theta) - coord.u_coord) / sin(theta);
    }

    return param;
}
/* 
 * returns the determinant of | a  b |  
 *                            | c  d |
 **/
double determinant(double a, double b, double c, double d){
    return(a*d - b*c);
}

/**
 * Checking the validity of the arguments (via commandline) passed. (Called by main())
 *
 * @param argv[1] = nestCoordsFileName
 *       argv[2] = numDirections
 *       argv[3] = SAMPLESIZE (i.e. number of samples after burnin)
 *       argv[4] = IGNORESAMPLE (i.e. burn in)
 *       argv[5] = outputFileName
 *
 * @param argc (number of items on command line: needs to be equal to 6)
 * @param **argv (double pointer?)
 * @pre
 * @post
 */
void handleargs(int argc, char **argv){
    if (argc!=6){
        printf("Invalid commandline.\n");
        exit(1);
    }
    //since nesCoordsFileName[100], the programs terminates if file name length>100
    if (strlen(argv[1]) > 100){
        printf("Nest coords file name >100.\n");
        exit(1);
    }
    //Copy string, from argv[1] into the array nestCordsFileName
    strcpy(nestCoordsFileName, argv[1]); 

    //Read formatted data from string
    sscanf(argv[2], "%d", &numDirections);
    sscanf(argv[3], "%d", &SAMPLESIZE);
    sscanf(argv[4], "%d", &IGNORESAMPLE);

    if (strlen(argv[5]) > 100){
        printf("output file name >100.\n");
        exit(1);
    }
    strcpy(outputFileName, argv[5]);
   


}


double *comparelen;
int comparelength(const void *e1,const void *e2)
{
  //Sort by decreasing length, then by increasing p1, then increasing p2.
  struct edge *edge1 = (struct edge*)e1;
  struct edge *edge2 = (struct edge*)e2;
  
  double len1 = comparelen[edge1->t1];
  double len2 = comparelen[edge2->t1];
  
  if(len1 > len2) return -1;
  else if(len1 < len2) return 1;
  else if(edge1->p1 == edge2->p1) return edge1->p2 - edge2->p2;
  else return edge1->p1 - edge2->p1;
}


int comparearea(const void *e1,const void *e2)
{
  struct poly *poly1 = (struct poly*)e1;
  struct poly *poly2 = (struct poly*)e2;
  
  if(poly1->area < poly2->area) return -1;
  else if(poly1->area > poly2->area) return 1;
  else return 0;
}


float *boundarypoints(float *points,int numpoints,int *numboundarypoints,double *area)
{
  unsigned int *indexlist;
  int numTriangleVertices;
  int vnum;

  unsigned int *BuildTriangleIndexList (void *pointList, float factor, int numberOfInputPoints, int numDimensions, int clockwise, int *numTriangleVertices );
  
  indexlist = BuildTriangleIndexList (points, 1.0, numpoints, 2, 1, &numTriangleVertices);
  
  //Build list of edges
  struct edge *edgelist = (struct edge*)malloc(numTriangleVertices*sizeof(struct edge));
  double *len = (double*)malloc(numTriangleVertices*sizeof(double));
  for(vnum=0; vnum < numTriangleVertices; vnum++){
    edgelist[vnum].p1 = indexlist[vnum];
    if(vnum%3 == 2) edgelist[vnum].p2 = indexlist[vnum-2];
    else edgelist[vnum].p2 = indexlist[vnum+1];
    if(edgelist[vnum].p2 < edgelist[vnum].p1){
      //Swap vertices
      edgelist[vnum].p1 = edgelist[vnum].p2;
      edgelist[vnum].p2 = indexlist[vnum];
    }
    edgelist[vnum].t1 = vnum;
    
    double x1 = points[2*(edgelist[vnum].p1)];
    double y1 = points[2*(edgelist[vnum].p1)+1];
    double x2 = points[2*(edgelist[vnum].p2)];
    double y2 = points[2*(edgelist[vnum].p2)+1];
    len[vnum] = sqrt((x1-x2)*(x1-x2) + (y1-y2)*(y1-y2));
  }
  
  //sort by length
  comparelen = len;
  qsort(edgelist,numTriangleVertices,sizeof(struct edge),&comparelength);
  
  //identify internal and external edges
  int *eext = (int*)malloc(numTriangleVertices*sizeof(int));
  int *partner = (int*)malloc(numTriangleVertices*sizeof(int));
  eext[edgelist[numTriangleVertices-1].t1] = 1; //Shortest edge is external by default
  partner[edgelist[numTriangleVertices-1].t1] = -1;
  for(vnum=0; vnum < numTriangleVertices-1; vnum++){
    if((edgelist[vnum].p1 == edgelist[vnum+1].p1) && (edgelist[vnum].p2 == edgelist[vnum+1].p2))
    {
      eext[edgelist[vnum].t1] = eext[edgelist[vnum+1].t1] = 0;
      partner[edgelist[vnum].t1] = edgelist[vnum+1].t1;
      partner[edgelist[vnum+1].t1] = edgelist[vnum].t1;
      vnum++;
    }
    else{
      eext[edgelist[vnum].t1] = 1;
      partner[edgelist[vnum].t1] = -1;
    }
  }
  
  //identify internal and external points
  int *pext = (int*)malloc(numpoints*sizeof(int));
  for(vnum=0;vnum<numpoints;vnum++) pext[vnum]=0;
  for(vnum=0;vnum<numTriangleVertices;vnum++){
    if(eext[vnum]){
      if(vnum%3 == 2) pext[indexlist[vnum]]=pext[indexlist[vnum-2]]=1;
      else pext[indexlist[vnum]]=pext[indexlist[vnum+1]]=1;
    }
  }
  
  //Remove external edges that can be removed
  int *removed = (int*)malloc(numTriangleVertices*sizeof(int));
  for(vnum=0;vnum<numTriangleVertices;vnum++) removed[vnum]=0;
  for(vnum=0;vnum<numTriangleVertices;vnum++){
    int edge = edgelist[vnum].t1;
    int edgeA = edge+1;
    if(edgeA%3 == 0) edgeA-=3;
    int edgeB = edgeA+1;
    if(edgeB%3 == 0) edgeB-=3;
    int vert = indexlist[edgeB];
    if(!removed[edge] && len[edge]>MINLEN && eext[edge] && !eext[edgeA] && !eext[edgeB] && !pext[vert]){
      removed[edge] = removed[edgeA] = removed[edgeB] = 1;
      eext[partner[edgeA]] = eext[partner[edgeB]] = 1;
      pext[vert] = 1;
      vnum = -1; //Very inefficient - returns to the beginning after each edge removal
    }
  }
  
  //Compute area
  *area=0.0;
  for(vnum=0;vnum<numTriangleVertices;vnum+=3){
    if(removed[vnum]) continue;
    double a = len[vnum];
    double b = len[vnum+1];
    double c = len[vnum+2];
    double p = (a + b + c)/2.0f;
    *area += sqrt(p*(p-a)*(p-b)*(p-c));  //Heron's formula
  }
  
  //Construct clockwise simple polygon
  for(vnum=0; vnum<numTriangleVertices; vnum++){
    if(eext[vnum] && !removed[vnum]) break; //First external edge
  }
  
  int next = vnum;
  *numboundarypoints = 0;
  float *boundary = (float*)malloc(2*numpoints*sizeof(float));
  
  do{
    boundary[2*(*numboundarypoints)] = points[2*indexlist[next]];
    boundary[2*(*numboundarypoints)+1] = points[2*indexlist[next]+1];
    (*numboundarypoints)++;
    next++;
    if(next%3 == 0) next -= 3;
    
    while(!eext[next]){
      next = partner[next];
      next++;
      if(next%3 == 0) next -= 3;
    }
  }
  while(next!=vnum);

  free(indexlist);
  free(edgelist);
  free(len);
  free(eext);
  free(pext);
  free(removed);
  free(partner);

  return boundary;
}


/*
 * Runs the code for finding north/south etc. extreme median boundaries
 * The points will eventually form a CONVEX polygon
 * When this is called, the burn-in samples should already have been read over
 *
 * @param
 *
 * @pre
 */
void simpleBoundaries(FILE *nestCoords, Location *LocationList){
    int time; //counter to iterate through the time intervals;
    int sample;//counter to iterate through the (non burn in) samples;
  
  Array2D<struct poly> DataPoints(NUMTIME,SAMPLESIZE);
  
  
    for (sample = 0; sample < SAMPLESIZE; sample++){
        int maxNumNests = readSampleHeader(nestCoords);
        int actualNumNests = readSample(nestCoords, maxNumNests, LocationList);

        for (time = FIRSTTIME; time < NUMTIME; time++){
            //geting the coordinates list for this sample and time
            float *currentList = (float*)malloc(2*actualNumNests*sizeof(float));
                    
            //we don't include nests that don't exist during this time,
            //hence yet another counter
            int listSizeGivenTime = 0;
            for (int nestnum = 0; nestnum < actualNumNests; nestnum++){
                if (LocationList[nestnum].time_start <= time
                    &&  LocationList[nestnum].time_end >= time){
                    //the nest existed during time
                    currentList[2*listSizeGivenTime] = LocationList[nestnum].x_coord - 480000;
                    currentList[2*listSizeGivenTime + 1] = LocationList[nestnum].y_coord - 6940000;
                    listSizeGivenTime++;
                }
            }

            int numboundarypoints;
            float *boundary = (float*)malloc(2*listSizeGivenTime*sizeof(float));
            double area;
          
            boundary = boundarypoints(currentList,listSizeGivenTime,&numboundarypoints,&area);
                    
            //Store external points
            DataPoints(time,sample).numpoints = numboundarypoints;
            DataPoints(time,sample).points = (float*)malloc(2*numboundarypoints*sizeof(float));
            for(int i=0;i<2*numboundarypoints;i++){
              DataPoints(time,sample).points[i] = boundary[i];
            }
            DataPoints(time,sample).area = area;
          
            free(currentList);
            free(boundary);
        }
    }

    for (time = FIRSTTIME; time < NUMTIME; time++){
      qsort(DataPoints(time),SAMPLESIZE,sizeof(struct poly),&comparearea);

      //Combine boundary points
      for(int i=1;i<SAMPLESIZE;i++){
        float *boundary = (float*)malloc(2*(DataPoints(time,i-1).numpoints+DataPoints(time,i).numpoints)*sizeof(float));
        for(int j=0;j<2*DataPoints(time,i-1).numpoints;j++){
          boundary[j] = DataPoints(time,i-1).points[j];
        }
        for(int j=0;j<2*DataPoints(time,i).numpoints;j++){
          boundary[2*DataPoints(time,i-1).numpoints + j] = DataPoints(time,i).points[j];
        }
        free(DataPoints(time,i).points);
        DataPoints(time,i).points = boundarypoints(boundary,DataPoints(time,i-1).numpoints+DataPoints(time,i).numpoints,&(DataPoints(time,i).numpoints),&(DataPoints(time,i).area));
        free(boundary);
      }
      
      for(int i=0;i<SAMPLESIZE;i++){
        for(int j=0;j<DataPoints(time,i).numpoints;j++){
          printf("%lf,%lf,",DataPoints(time,i).points[2*j]+480000,DataPoints(time,i).points[2*j+1]+6940000);
        }
        printf("\nArea=%lf\n",DataPoints(time,i).area);
      }
    }
  
    FILE* output = fopen(outputFileName,"w");
    if (!output){
        printf("output file failed to open.\n");
        exit(1);
    }

    //required for the fixed background on animation
    fprintf(output, "square = Polygon[{{%d,%d},{%d,%d},{%d,%d},{%d,%d}}];\n",
                        MINX, MINY, MAXX, MINY, MAXX, MAXY, MINX, MAXY);
    

    for (time = FIRSTTIME; time < NUMTIME; time++){
//        fprintf(output, "t%dcoord = Polygon[{", time);
//        for (rotation = 0; rotation < actualMedianPoints; rotation++){
//            fprintf(output, "{%f,%f}", medianPoints(time, rotation).u_coord, medianPoints(time, rotation).v_coord);
//            if (rotation != actualMedianPoints - 1){
//                fprintf(output, ",");
//            }
//        }
//        fprintf(output, "}];\n");
//
//
//        fprintf(output, "t%dlower = Polygon[{", time);
//        for (rotation = 0; rotation < actualLowerPoints; rotation++){
//            fprintf(output, "{%f,%f}", LowerPoints(time, rotation).u_coord, LowerPoints(time, rotation).v_coord);
//            if (rotation != actualLowerPoints - 1){
//                fprintf(output, ",");
//            }
//        }
//        fprintf(output, "}];\n");
//
//        fprintf(output, "t%dupper = Polygon[{", time);
//        for (rotation = 0; rotation < actualUpperPoints; rotation++){
//            fprintf(output, "{%f,%f}", UpperPoints(time, rotation).u_coord, UpperPoints(time, rotation).v_coord);
//            if (rotation != actualUpperPoints - 1){
//                fprintf(output, ",");
//            }
//        }
//        fprintf(output, "}];\n");
//        
//        fprintf(output, "TIME%d = Graphics[{{Opacity[0],square},{Yellow,t%dupper},{Orange,t%dcoord},{Red,t%dlower}},Axes -> True];\n", time, time,time,time);
    }

    //completing rest of output
    fprintf(output , "list = {");
    for (int time = FIRSTTIME; time < NUMTIME; time++){
        fprintf(output, "TIME%d", time);
        if (time != NUMTIME - 1){
            fprintf(output, ",");
        }
    }
    fprintf(output, "};\n");
    fprintf(output, "ListAnimate[list,AnimationRunning -> False]\n");
    fprintf(output, "Export[\"%s_convex_%ddirc_%dsamples.swf\", ListAnimate[list, AnimationRunning->False], \"ControlAppearance\"->None, \"Scalable\"->True, \"FrameRate\" -> 8, \"AnimationDuration\" -> 24]", outputFileName, numDirections, SAMPLESIZE);

    fclose(output);
}

/*
 * Convex polygon, ignoring exterior lines
 * @param argv[1] = nestCoordsFileName
 *        argv[2] = numDirections
 *        argv[3] = SAMPLESIZE (i.e. number of samples after burnin)
 *        argv[4] = IGNORESAMPLE (i.e. burn in)
 *        argv[5] = mathematica-styled output file name
 */
int main(int argc, char **argv) {
    handleargs(argc, argv);
    //angle increment, computed using numDirections
    angleIncrement = 2 * PI / numDirections;

    //things that always happen regardless of relative/radial Boundaries
    Location *LocationList = (Location*)malloc(MAXNUMNESTS*sizeof(Location));
    readFixed(LocationList); //happens always

    FILE *nestCoords;//reading in the nest Coordinates
    nestCoords = fopen(nestCoordsFileName, "r");
    if (!nestCoords){
        printf("Could not open file %s.\n", nestCoordsFileName);
        exit(1);
    }
    else{
        printf("Opened file %s.\n", nestCoordsFileName);
    }

    printf("Code version edited on 30-6-2015 2:41PM.\n");

    printf("///////////////////////Ignored Samples///////////////////////\n");
    for (int i = 0; i < IGNORESAMPLE; i++){
        ignoreSample(nestCoords);
    }
    printf("///////////////////////Finished Ignoring///////////////////////\n");

    simpleBoundaries(nestCoords, LocationList);
    
    free(LocationList);
    fclose(nestCoords);
}
