/*******************************************************************************
*
* ants_map
* Code written by Jonathan Keith
* July 2015 - tidied up March 2017
*
********************************************************************************
*/

#include <iostream>
#include <string>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "PointStruct.h" //for Point arrays
#include "ReadData.h"

#define MINLEN 10000
int SAMPLESIZE; // after burn in
int IGNORESAMPLE; //burn in


char nestCoordsFileName[100];
char outputFileName[100];

/*
* Defines a wrapper class for a 2D array, implemented as a 1D array
*/
template <typename T>
class Array2D{
    // private data members
private:
    int rowNum;
    int columnNum;
    T* Matrix;

public:
    /*
    * Constructor
    */
    Array2D(int row, int column){
        if (row > 0 && column > 0){
            rowNum = row;
            columnNum = column;
            Matrix = new T[rowNum *columnNum + columnNum];
        }

    }

    /*
    * Deconstructor
    */
    ~Array2D(){
        delete[] Matrix;
    }

    /*
    * Returns pointer to item at Matrx[x][y][0]
    * i.e, an "array" of size depthNum
    */
    T* operator () (int x){
        return &Matrix[x*columnNum + 0];
    }

    /*
    * Returns item at Matrix[x][y][z]
    */
    const T& operator () (int x, int y) const{
        return Matrix[x*columnNum + y];
    }
    T& operator () (int x, int y){
        return Matrix[x*columnNum + y];
    }

    int getRowNum(){
        return rowNum;
    }

    int getColumnNum(){
        return columnNum;
    }

};


struct edge{
  int p1; //End point 1
  int p2; //End point 2
  int t1; //occurence of edge in indexlist
};


/**
 * Checking the validity of the arguments (via commandline) passed. (Called by main())
 *
 * @param argv[1] = nestCoordsFileName
 *       argv[2] = SAMPLESIZE (i.e. number of samples after burnin)
 *       argv[3] = IGNORESAMPLE (i.e. burn in)
 *       argv[4] = outputFileName
 *
 * @param argc (number of items on command line: needs to be equal to 5)
 */
void handleargs(int argc, char **argv){
    if (argc!=5){
        printf("Invalid command line.\n");
        exit(1);
    }
    //since nesCoordsFileName[100], the programs terminates if file name length>100
    if (strlen(argv[1]) > 100){
        printf("Nest coords file name >100.\n");
        exit(1);
    }
    //Copy string, from argv[1] into the array nestCordsFileName
    strcpy(nestCoordsFileName, argv[1]); 

    //Read formatted data from string
    sscanf(argv[2], "%d", &SAMPLESIZE);
    sscanf(argv[3], "%d", &IGNORESAMPLE);

    if (strlen(argv[4]) > 100){
        printf("output file name >100.\n");
        exit(1);
    }
    strcpy(outputFileName, argv[4]);
}


double *comparelen;
int comparelength(const void *e1,const void *e2)
{
  //Sort by decreasing length, then by increasing p1, then increasing p2.
  struct edge *edge1 = (struct edge*)e1;
  struct edge *edge2 = (struct edge*)e2;
  
  double len1 = comparelen[edge1->t1];
  double len2 = comparelen[edge2->t1];
  
  if(len1 > len2) return -1;
  else if(len1 < len2) return 1;
  else if(edge1->p1 == edge2->p1) return edge1->p2 - edge2->p2;
  else return edge1->p1 - edge2->p1;
}


double *boundarypoints(float *points,int numpoints,int *numboundarypoints)
{
  unsigned int *indexlist;
  int numTriangleVertices;
  int vnum;

  unsigned int *BuildTriangleIndexList (void *pointList, float factor, int numberOfInputPoints, int numDimensions, int clockwise, int *numTriangleVertices );
  
  indexlist = BuildTriangleIndexList (points, 1.0, numpoints, 2, 1, &numTriangleVertices);
  
  //Build list of edges
  struct edge *edgelist = (struct edge*)malloc(numTriangleVertices*sizeof(struct edge));
  double *len = (double*)malloc(numTriangleVertices*sizeof(double));
  for(vnum=0; vnum < numTriangleVertices; vnum++){
    edgelist[vnum].p1 = indexlist[vnum];
    if(vnum%3 == 2) edgelist[vnum].p2 = indexlist[vnum-2];
    else edgelist[vnum].p2 = indexlist[vnum+1];
    if(edgelist[vnum].p2 < edgelist[vnum].p1){
      //Swap vertices
      edgelist[vnum].p1 = edgelist[vnum].p2;
      edgelist[vnum].p2 = indexlist[vnum];
    }
    edgelist[vnum].t1 = vnum;
    
    double x1 = points[2*(edgelist[vnum].p1)];
    double y1 = points[2*(edgelist[vnum].p1)+1];
    double x2 = points[2*(edgelist[vnum].p2)];
    double y2 = points[2*(edgelist[vnum].p2)+1];
    len[vnum] = sqrt((x1-x2)*(x1-x2) + (y1-y2)*(y1-y2));
  }
  
  //sort by length
  comparelen = len;
  qsort(edgelist,numTriangleVertices,sizeof(struct edge),&comparelength);
  
  //identify internal and external edges
  int *eext = (int*)malloc(numTriangleVertices*sizeof(int));
  int *partner = (int*)malloc(numTriangleVertices*sizeof(int));
  eext[edgelist[numTriangleVertices-1].t1] = 1; //Shortest edge is external by default
  partner[edgelist[numTriangleVertices-1].t1] = -1;
  for(vnum=0; vnum < numTriangleVertices-1; vnum++){
    if((edgelist[vnum].p1 == edgelist[vnum+1].p1) && (edgelist[vnum].p2 == edgelist[vnum+1].p2))
    {
      eext[edgelist[vnum].t1] = eext[edgelist[vnum+1].t1] = 0;
      partner[edgelist[vnum].t1] = edgelist[vnum+1].t1;
      partner[edgelist[vnum+1].t1] = edgelist[vnum].t1;
      vnum++;
    }
    else{
      eext[edgelist[vnum].t1] = 1;
      partner[edgelist[vnum].t1] = -1;
    }
  }
  
  //identify internal and external points
  int *pext = (int*)malloc(numpoints*sizeof(int));
  for(vnum=0;vnum<numpoints;vnum++) pext[vnum]=0;
  for(vnum=0;vnum<numTriangleVertices;vnum++){
    if(eext[vnum]){
      if(vnum%3 == 2) pext[indexlist[vnum]]=pext[indexlist[vnum-2]]=1;
      else pext[indexlist[vnum]]=pext[indexlist[vnum+1]]=1;
    }
  }
  
  //Remove external edges that can be removed
  int *removed = (int*)malloc(numTriangleVertices*sizeof(int));
  for(vnum=0;vnum<numTriangleVertices;vnum++) removed[vnum]=0;
  for(vnum=0;vnum<numTriangleVertices;vnum++){
    int edge = edgelist[vnum].t1;
    int edgeA = edge+1;
    if(edgeA%3 == 0) edgeA-=3;
    int edgeB = edgeA+1;
    if(edgeB%3 == 0) edgeB-=3;
    int vert = indexlist[edgeB];
    if(!removed[edge] && len[edge]>MINLEN && eext[edge] && !eext[edgeA] && !eext[edgeB] && !pext[vert]){
      removed[edge] = removed[edgeA] = removed[edgeB] = 1;
      eext[partner[edgeA]] = eext[partner[edgeB]] = 1;
      pext[vert] = 1;
      vnum = -1; //Very inefficient - returns to the beginning after each edge removal
    }
  }
  
  //Construct clockwise simple polygon
  for(vnum=0; vnum<numTriangleVertices; vnum++){
    if(eext[vnum] && !removed[vnum]) break; //First external edge
  }
  
  int next = vnum;
  *numboundarypoints = 0;
  double *boundary = (double*)malloc(2*numpoints*sizeof(double));
  
  do{
    boundary[2*(*numboundarypoints)] = points[2*indexlist[next]];
    boundary[2*(*numboundarypoints)+1] = points[2*indexlist[next]+1];
    (*numboundarypoints)++;
    next++;
    if(next%3 == 0) next -= 3;
    
    while(!eext[next]){
      next = partner[next];
      next++;
      if(next%3 == 0) next -= 3;
    }
  }
  while(next!=vnum);

  free(indexlist);
  free(edgelist);
  free(len);
  free(eext);
  free(pext);
  free(removed);
  free(partner);

  return boundary;
}


int doublecompare(const void *e1,const void *e2)
{
  double *x1 = (double*)e1;
  double *x2 = (double*)e2;
  
  if(*x1 < *x2) return -1;
  else if(*x1 > *x2) return 1;
  return 0;
}


void updategrid(double *boundary,int numboundarypoints,Array2D<long> &grid,int xstart,int ystart,int gridsize)
{
  int bottomrow,toprow;
  
  //Get lists of crossovers
  Array2D<double> crossovers(grid.getRowNum(),numboundarypoints);
  int *numcrossovers = (int*)malloc(grid.getRowNum()*sizeof(int));
  for(int i=0;i<grid.getRowNum();i++) numcrossovers[i] = 0;

  for(int i=0; i<numboundarypoints;i++){
    double x1 = boundary[2*i];
    double y1 = boundary[2*i+1];
    double x2,y2;
    if(i == numboundarypoints-1){
      x2 = boundary[0];
      y2 = boundary[1];
    }
    else{
      x2 = boundary[2*(i+1)];
      y2 = boundary[2*(i+1)+1];
    }
    if(y1==y2) continue;

    if(y1<y2){
      bottomrow = (int)ceil((y1 - ystart)/gridsize);
      if(ystart + (bottomrow-1)*gridsize >= y1) bottomrow--;
      toprow = (int)floor((y2 - ystart)/gridsize);
      if(ystart + toprow*gridsize == y2) toprow--;
      else if(ystart + (toprow+1)*gridsize < y2) toprow++;
    }
    else{
      bottomrow = (int)ceil((y2 - ystart)/gridsize);
      if(ystart + (bottomrow-1)*gridsize >= y2) bottomrow--;
      toprow = (int)floor((y1 - ystart)/gridsize);
      if(ystart + toprow*gridsize == y1) toprow--;
      else if(ystart + (toprow+1)*gridsize < y1) toprow++;
    }

    double m = (x2 - x1) / (y2 - y1);
    for(int j=bottomrow;j<=toprow;j++){
      double y = ystart + j*gridsize;
      crossovers(j,numcrossovers[j]) = x1 + m*(y - y1);
      numcrossovers[j]++;
    }
  }

  //Get bounding box of boundary
  double left,right,top,bottom;
  left = right = boundary[0];
  top = bottom = boundary[1];
  for(int i=1;i<numboundarypoints;i++){
    if(boundary[2*i] < left) left = boundary[2*i];
    else if(boundary[2*i] > right) right = boundary[2*i];
    if(boundary[2*i+1] < bottom) bottom = boundary[2*i+1];
    else if(boundary[2*i+1] > top) top = boundary[2*i+1];
  }
  
  //Sort crossovers
  bottomrow = (int)floor((bottom - ystart)/gridsize);
  toprow = (int)ceil((top - ystart)/gridsize);
  for(int i=bottomrow;i<=toprow;i++){
    if(numcrossovers[i]>0) qsort(crossovers(i), numcrossovers[i], sizeof(double), &doublecompare);
  }
  
  //Increment grid
  int leftcol = (int)ceilf((left - xstart)/gridsize);
  int rightcol = (int)floorf((right - xstart)/gridsize);
  for(int i=bottomrow;i<=toprow;i++){
    int count = 0;
    for(int j=leftcol;j<=rightcol;j++){
      double x = xstart + j*gridsize;
      while(count<numcrossovers[i] && (x > crossovers(i,count) || (x == crossovers(i,count) && count%2 == 0))) count++;
      if(count%2 == 1) grid(j,i)++;
    }
  }
  
  free(numcrossovers);
}


/*
 * Runs the code for finding north/south etc. extreme median boundaries
 * The points will eventually form a CONVEX polygon
 * When this is called, the burn-in samples should already have been read over
 *
 * @param
 *
 * @pre
 */
void simpleBoundaries(FILE *nestCoords, Location *LocationList){
    int time = NUMTIME-1;
    int sample;//counter to iterate through the (non burn in) samples;
  
    int gridsize = 100;
    int xstart = (MINX/gridsize)*gridsize;
    if(xstart < MINX) xstart=(MINX/gridsize+1)*gridsize;
    int ystart = (MINY/gridsize)*gridsize;
    if(ystart < MINY) ystart = (MINY/gridsize+1)*gridsize;
    int xwidth = (MAXX - xstart + 1)/gridsize;
    int ywidth = (MAXY - ystart + 1)/gridsize;
    int xoffset = 480000;
    int yoffset = 6930000;
    xstart -= xoffset;
    ystart -= yoffset;

    Array2D<long> grid(xwidth,ywidth);
    for(int i=0;i<xwidth;i++){
      for(int j=0;j<ywidth;j++){
        grid(i,j) = 0;
      }
    }

    for (sample = 0; sample < SAMPLESIZE; sample++){
        int maxNumNests = readSampleHeader(nestCoords);
        int actualNumNests = readSample(nestCoords, maxNumNests, LocationList);
        //geting the coordinates list for this sample and time
        float *currentList = (float*)malloc(2*actualNumNests*sizeof(float));

        //we don't include nests that don't exist during this time,
        //hence yet another counter
        int listSizeGivenTime = 0;
        for (int nestnum = 0; nestnum < actualNumNests; nestnum++){
          if(LocationList[nestnum].x_coord<MINX || LocationList[nestnum].x_coord > MAXX ||
             LocationList[nestnum].y_coord<MINY || LocationList[nestnum].y_coord>MAXY){
            printf("*!*!* %lf %lf\n",LocationList[nestnum].x_coord,LocationList[nestnum].y_coord);
          }
            if (LocationList[nestnum].time_start <= time
                &&  LocationList[nestnum].time_end >= time){
                //the nest existed during time
                currentList[2*listSizeGivenTime] = (float)(LocationList[nestnum].x_coord - xoffset);
                currentList[2*listSizeGivenTime + 1] = (float)(LocationList[nestnum].y_coord - yoffset);
                listSizeGivenTime++;
            }
        }

        int numboundarypoints;
        double *boundary = boundarypoints(currentList,listSizeGivenTime,&numboundarypoints);
        free(currentList);

        updategrid(boundary,numboundarypoints,grid,xstart,ystart,gridsize);
      
        free(boundary);
    }

    //Create polygons
    double threshvals[7] = {0.999,0.99,0.975,0.75,0.5,0.25,0.025};
    double *boundary[7];
    int boundarylen[7];
    float *list = (float*)malloc(2*xwidth*ywidth*sizeof(float));
  
    for(int poly=0;poly<7;poly++){
      int numboundarypoints = 0;
      int thresh = (int)floor((1.0-threshvals[poly])*SAMPLESIZE);
      for(int i=0;i<xwidth;i++){
        for(int j=0;j<ywidth;j++){
          if((grid(i,j) > thresh) && (i==0 || i==xwidth-1 || j==0 || j==ywidth-1 ||
               (grid(i-1,j) <= thresh) || (grid(i+1,j) <= thresh) ||
               (grid(i,j-1) <= thresh) || (grid(i,j+1) <= thresh))){
            list[2*numboundarypoints] = (float)(xstart + i*gridsize);
            list[2*numboundarypoints+1] = (float)(ystart + j*gridsize);
            numboundarypoints++;
          }
        }
      }
    
      boundary[poly] = boundarypoints(list,numboundarypoints,boundarylen+poly);
    }
    free(list);

    //Output
    FILE* output = fopen(outputFileName,"w");
    if (!output){
        printf("output file failed to open.\n");
        exit(1);
    }

    fprintf(output, "square = Polygon[{{%d,%d},{%d,%d},{%d,%d},{%d,%d}}];\n",
                        MINX, MINY, MAXX, MINY, MAXX, MAXY, MINX, MAXY);
    

    for (int poly = 0; poly < 7; poly++){
        fprintf(output, "t%d-p%dcoord = Polygon[{",time,(int)(1000*threshvals[poly]));
        for (int i = 0; i < boundarylen[poly]; i++){
            fprintf(output, "{%f,%f}", boundary[poly][2*i]+xoffset,boundary[poly][2*i+1]+yoffset);
            if (i != boundarylen[poly] - 1){
                fprintf(output, ",");
            }
        }
        fprintf(output, "}];\n");
    }

    fclose(output);
  
    for(int poly=0;poly<7;poly++){
      free(boundary[poly]);
    }
}

int main(int argc, char **argv) {
    handleargs(argc, argv);

    Location *LocationList = (Location*)malloc(MAXNUMNESTS*sizeof(Location));
    readFixed(LocationList); 

    FILE *nestCoords;//reading in the nest Coordinates
    nestCoords = fopen(nestCoordsFileName, "r");
    if (!nestCoords){
        printf("Could not open file %s.\n", nestCoordsFileName);
        exit(1);
    }
    else{
        printf("Opened file %s.\n", nestCoordsFileName);
    }

    printf("///////////////////////Ignored Samples///////////////////////\n");
    for (int i = 0; i < IGNORESAMPLE; i++){
        ignoreSample(nestCoords);
    }
    printf("///////////////////////Finished Ignoring///////////////////////\n");

    simpleBoundaries(nestCoords, LocationList);
    
    free(LocationList);
    fclose(nestCoords);
}
